#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <drcom.h>

#include "config.h"
#include "drcom_types.h"
#include "client_daemon.h"
#include "daemon.h"

void msgprint(const char *);
void report_result(int, int, int);
void daemon_stopthreads(int);
void * daemon_watchport(void *);
void * daemon_keepalive(void *);
void module_start_auth(struct drcom_handle *);
void module_stop_auth(void);

/*
   0: idle
   1: logged in
   2: busy
*/
static int status = 0;
static pthread_t th_watchport = 0, th_keepalive = 0;

void drcomcd_daemon(int s)
{
  int s2, r;
#if 0
  int un_len;
  struct sockaddr_un un_client;
#endif
  struct drcom_handle *h;
  unsigned char buf1[DRCOMCD_BUF_LEN];
  struct drcomcd_hdr *cd_hdr = (struct drcomcd_hdr *) buf1;
  struct drcomcd_login *cd_login = (struct drcomcd_login *) (buf1 + sizeof(*cd_hdr));
  struct drcomcd_logout *cd_logout = (struct drcomcd_logout *) (buf1 + sizeof(*cd_hdr));
  struct drcomcd_passwd *cd_passwd = (struct drcomcd_passwd *) (buf1 + sizeof(*cd_hdr));

  /* Set up the signal handler early */
#ifdef DRCOM_DEBUG
  fprintf(stderr, "%10u DEBUG daemon: Setting up signal handler...\n", (unsigned) time(NULL));
#endif
  signal(SIGINT, daemon_stopthreads);

  /* Initialize the handle for the lifetime of the daemon */
#ifdef DRCOM_DEBUG
  fprintf(stderr, "%10u DEBUG daemon: Initializing drcom handle...\n", (unsigned) time(NULL));
#endif
  h = drcom_create_handle();
  drcom_init(h, &msgprint);

#if 0
  /* Either un_len be set to sizeof(un_client), or don't use it altogether */
  memset(&un_client, 0, sizeof(un_client));
  un_len = sizeof(un_client);
#endif

  while (1)
  {
#ifdef DRCOM_DEBUG
    fprintf(stderr, "%10u DEBUG daemon: Waiting for connection...\n", (unsigned) time(NULL));
#endif
#if 0
    s2 = accept(s, (struct sockaddr *) &un_client, (socklen_t *) &un_len);
#else
    s2 = accept(s, NULL, NULL);
#endif
    if (s2 == -1)
    {
      perror("daemon: accept");
      continue;
    }
#ifdef DRCOM_DEBUG
    fprintf(stderr, "%10u DEBUG daemon: accept()'ed connection, s2=%d\n", (unsigned) time(NULL), s2);
#endif

    /*
      Abnormal values:
      r < 0: an error occurred
      r == 0: "peer has performed an orderly shutdown"
      cd_hdr->signature != DRCOM_SIGNATURE: who sent this message?
    */
#ifdef DRCOM_DEBUG
    fprintf(stderr, "%10u DEBUG daemon: Waiting for request...\n", (unsigned) time(NULL));
#endif
    r = recv(s2, buf1, DRCOMCD_BUF_LEN, 0);
    if (r < 0)
      perror("daemon: recv");
    if (r < 1 || cd_hdr->signature != DRCOM_SIGNATURE)
    {
#ifdef DRCOM_DEBUG
      fprintf(stderr, "%10u DEBUG daemon: r=%d signature=0x%08x\n", (unsigned) time(NULL), r, cd_hdr->signature);
#endif
      fprintf(stderr, "Unknown signature\n");
      close(s2);
      continue;
    }

    switch (cd_hdr->type)
    {
      case DRCOMCD_LOGIN:
        if (status == 0)
        {
          status = 2;
#ifdef DRCOM_DEBUG
          fprintf(stderr, "%10u DEBUG daemon: Trying to login...\n", (unsigned) time(NULL));
#endif
          r = drcom_login(h, cd_login->timeout);
          status = (!r) ? 1 : 0;
#ifdef DRCOM_DEBUG
          fprintf(stderr, "%10u DEBUG daemon: After login, r=%d\n", (unsigned) time(NULL), r);
#endif
        }
        else
        {
          report_result(s2, DRCOMCD_FAILURE, DRCOMCD_REASON_BUSY);
          continue;
        }

        /* Result-dependent task */
        if (status == 1)
        {
          module_start_auth(h);
          pthread_create(&th_watchport, NULL, daemon_watchport, h);
          pthread_create(&th_keepalive, NULL, daemon_keepalive, h);
        }

        break;

      case DRCOMCD_LOGOUT:
        if (status == 1)
        {
          status = 2;
          /* Stop the threads here, since they might interfere with
             the logout process */
          module_stop_auth();
          pthread_kill(th_keepalive, SIGINT);
          pthread_kill(th_watchport, SIGINT);
          /* Now try to log out */
#ifdef DRCOM_DEBUG
          fprintf(stderr, "%10u DEBUG daemon: Trying to logout...\n", (unsigned) time(NULL));
#endif
          r = drcom_logout(h, cd_logout->timeout);
#ifdef DRCOM_DEBUG
          fprintf(stderr, "%10u DEBUG daemon: After logout, r=%d\n", (unsigned) time(NULL), r);
#endif
          /* If logout failed, that means we are still logged in,
             so re-create the threads and continue authentication */
          if (r)
          {
            module_start_auth(h);
            pthread_create(&th_watchport, NULL, daemon_watchport, h);
            pthread_create(&th_keepalive, NULL, daemon_keepalive, h);
          }
          status = (!r) ? 0 : 1;
        }
        else
        {
          report_result(s2, DRCOMCD_FAILURE, DRCOMCD_REASON_BUSY);
          continue;
        }

        break;

      case DRCOMCD_PASSWD:
        if (status == 0)
        {
          status = 2;
#ifdef DRCOM_DEBUG
          fprintf(stderr, "%10u DEBUG daemon: Trying to change password...\n", (unsigned) time(NULL));
#endif
          r = drcom_passwd(h, cd_passwd->newpasswd, cd_passwd->timeout);
#ifdef DRCOM_DEBUG
          fprintf(stderr, "%10u DEBUG daemon: After password change, r=%d\n", (unsigned) time(NULL), r);
#endif
          status = 0;
        }
        else
        {
          report_result(s2, DRCOMCD_FAILURE, DRCOMCD_REASON_BUSY);
          continue;
        }

        break;

      default: continue; break;
    }

    if (r)
      report_result(s2, DRCOMCD_FAILURE, DRCOMCD_REASON_NO_REASON);
    else
      report_result(s2, DRCOMCD_SUCCESS, DRCOMCD_REASON_NO_REASON);
  }

  return;
}

void msgprint(const char *s)
{
  fprintf(stderr, "%s\n", s);
  return;
}

void report_result(int s2, int result, int reason)
{
  unsigned char buf2[DRCOMCD_BUF_LEN];
  struct drcomcd_hdr *dc_hdr = (struct drcomcd_hdr *) buf2;
  struct drcomcd_result *dc_result = (struct drcomcd_result *) (buf2 + sizeof(*dc_hdr));
  int r;

  dc_hdr->signature = DRCOM_SIGNATURE;
  dc_hdr->type = result;
  dc_result->reason = reason;

#ifdef DRCOM_DEBUG
  fprintf(stderr, "%10u DEBUG daemon: Reporting result...\n", (unsigned) time(NULL));
#endif
  r = send(s2, buf2, sizeof(*dc_hdr) + sizeof(*dc_result), 0);
  if (r < 0)
    perror("daemon: send");

  close(s2);

  return;
}

void daemon_stopthreads(int signum)
{
  pthread_t self;

  if (signum != SIGINT) return;

  self = pthread_self();
  if (self == th_watchport)
  {
    th_watchport = 0;
    pthread_exit(NULL);
  }
  else if (self == th_keepalive)
  {
    th_keepalive = 0;
    pthread_exit(NULL);
  }

  return;
}

void * daemon_watchport(void *arg)
{
  drcom_watchport((struct drcom_handle *) arg);
printf("watchport returns\n");
  return NULL;
}

void * daemon_keepalive(void *arg)
{
  sleep(9);
  drcom_keepalive((struct drcom_handle *) arg);
printf("keepalive returns\n");
  return NULL;
}

void module_start_auth(struct drcom_handle *h)
{
  struct drcom_status_data status_data;
  struct drcom_auth_data auth_data;
  struct drcom_iface_data iface_data;
  struct drcom_except_data except_data[MAX_EXCEPT_ITEMS];
  struct drcom_session_info *s;
  int authlen, r;
  FILE *f;

  s = drcom_get_session_info(h);
  authlen = drcom_get_authlen();

  /* struct drcom_auth is already packed, so no worries here */
  memcpy(auth_data.auth, s->auth, authlen);

  /* memset(drcom_iface_data, 0, 4 * sizeof(struct addrport)); */

  iface_data.hostip = s->hostip;
  iface_data.hostport = s->hostport;
  iface_data.servip = s->servip;
  iface_data.servport = s->servport;
  iface_data.dnsp = s->dnsp;
  iface_data.dnss = s->dnss;

  /* memset(drcom_except_data, 0, 20 * sizeof(struct addrmask)); */

  except_data[0].addr = 0;
  except_data[0].mask = 0;

  /* Actually write to the files */

  f = fopen(DRCOM_MODULE_AUTH, "w");
  if (!f)
  {
    perror("Writing authentication data");
    return;
  }
  r = fwrite(&auth_data, AUTH_LEN, 1, f);
  if (r == 0 && ferror(f))
  {
    perror("Writing to " DRCOM_MODULE_AUTH);
    return;
  }
  fclose(f);

  f = fopen(DRCOM_MODULE_IFACE, "w");
  if (!f)
  {
    perror("Writing interface info");
    return;
  }
  r = fwrite(&iface_data, IFACE_LEN, 1, f);
  if (r == 0 && ferror(f))
  {
    perror("Writing to " DRCOM_MODULE_IFACE);
    return;
  }
  fclose(f);

  f = fopen(DRCOM_MODULE_EXCEPT, "w");
  if (!f)
  {
    perror("Writing exceptions");
    return;
  }
  r = fwrite(&except_data, EXCEPT_LEN, MAX_EXCEPT_ITEMS, f);
  if (r == 0 && ferror(f))
  {
    perror("Writing to " DRCOM_MODULE_EXCEPT);
    return;
  }
  fclose(f);

  /* Now tell drcom.o to start authenticating */

  status_data.status = STATUS_LOGIN;

  f = fopen(DRCOM_MODULE_STATUS, "w");
  if (!f)
  {
    perror("Starting authentication");
    return;
  }
  r = fwrite(&status_data, STATUS_LEN, 1, f);
  if (r == 0 && ferror(f))
  {
    perror("Writing to " DRCOM_MODULE_STATUS);
    return;
  }
  fclose(f);

  return;
}

void module_stop_auth()
{
  struct drcom_status_data status_data;
  FILE *f;
  int r;

  status_data.status = STATUS_NOTLOGIN;

  f = fopen(DRCOM_MODULE_STATUS, "w");
  if (!f)
  {
    perror("Stopping authentication");
    return;
  }

  r = fwrite(&status_data, STATUS_LEN, 1, f);
  if (r == 0 && ferror(f))
  {
    perror("Writing to " DRCOM_MODULE_STATUS);
    return;
  }

  fprintf(stderr, "daemon: Stopped authentication\n");

  fclose(f);

  return;
}

