#include "drcom_linux.h"

/* for struct msghdr */
#include <linux/socket.h>
/* for struct proto */
#include <net/sock.h>
/* for struct proto_ops */
#include <linux/net.h>
/* for inet_stream_ops and inet_dgram_ops */
#include <net/inet_common.h>
/* for get_fs(), set_fs() */
#include <asm/uaccess.h>
/* for memcpy() */
#include <linux/string.h>
/* for E* */
#include <linux/errno.h>

/* include additional headers for 2.6.x */
#if (MAJOR_MINOR_IS(2,6))
/* for struct sockaddr_in */
#include <linux/in.h>
/* for inet_sk */
#include <linux/ip.h>
/* for TCP_* */      
#include <linux/tcp.h>
#endif

#include "config.h"
#include "drcom_types.h"
#include "proc.h"
#include "proto.h"

#if 0
struct drcom_status_data *status_data;
struct drcom_auth_data *auth_data;
struct drcom_iface_data *iface_data;
struct drcom_except_data *except_data;
#else
extern struct drcom_status_data *status_data;
extern struct drcom_auth_data *auth_data;
extern struct drcom_iface_data *iface_data;
extern struct drcom_except_data *except_data;
#endif

/* Use only for constants */

#define n_to_ip(a,b,c,d) \
          __constant_htonl( (((u_int32_t) a) << 24) \
                          + (((u_int32_t) b) << 16) \
                          + (((u_int32_t) c) << 8) \
                          +  ((u_int32_t) d) )

int init_proto(void);
int cleanup_proto(void);
int need_auth(u_int32_t, u_int32_t);
int drcom_udp_get_addrs(u_int32_t *, u_int32_t *, struct sock *, struct msghdr *);

/* These are replaced by drcom.o when it is loaded */

int (*system_inet_stream_ops_connect)(struct socket *, struct sockaddr *, int, int);
int drcom_inet_stream_ops_connect(struct socket *, struct sockaddr *, int, int);

#if (MAJOR_MINOR_IS(2,4))
int (*system_inet_dgram_ops_sendmsg)(struct socket *, struct msghdr *, int, struct scm_cookie *);
int drcom_inet_dgram_ops_sendmsg(struct socket *, struct msghdr *, int, struct scm_cookie *);
#else /* then it's a 2.6 kernel */
int (*system_inet_dgram_ops_sendmsg)(struct kiocb *iocb, struct socket *, struct msghdr *, size_t);
int drcom_inet_dgram_ops_sendmsg(struct kiocb *, struct socket *, struct msghdr *, size_t);
#endif

/* This is only occassionally replaced */

struct proto_ops *system_inet_stream_ops = &inet_stream_ops, drcom_inet_stream_ops;
#if (MAJOR_MINOR_IS(2,4))
int drcom_inet_stream_ops_sendmsg(struct socket *sock, struct msghdr *msg, int size, struct scm_cookie *scm);
#else /* then it's a 2.6 kernel */
int drcom_inet_stream_ops_sendmsg(struct kiocb *iocb, struct socket *sock, struct msghdr *msg, size_t size);
#endif

/* protocol handlers */

#if (MAJOR_MINOR_IS(2,4))
int drcom_inet_stream_ops_sendmsg(struct socket *sock, struct msghdr *msg, int size, struct scm_cookie *scm)
#else /* then it's a 2.6 kernel */
int drcom_inet_stream_ops_sendmsg(struct kiocb *iocb, struct socket *sock, struct msghdr *msg, size_t size)
#endif
{
  struct iovec iov, *old_iovec;
  int ret, n_iovec;
  mm_segment_t old_fs;

  /* We borrowed sock->ops, now it's time to return it ;) */
  sock->ops = system_inet_stream_ops;

  /* This is much simpler than udp */
  iov.iov_base = auth_data->auth;
  iov.iov_len = 16;

  /* Swap the iovecs (note: remember to swap msg_iovlen, too */
  n_iovec = msg->msg_iovlen;
  old_iovec = msg->msg_iov;

  /* Prepare for sending the auth packet */
  msg->msg_iovlen = 1;
  msg->msg_iov = &iov;

  /* Bypass address checking in copy_from_user() */
  old_fs = get_fs();
  set_fs(KERNEL_DS);
#if (MAJOR_MINOR_IS(2,4))
  ret = sock->ops->sendmsg(sock, msg, 16, scm);
#else /* then it's a 2.6 kernel */
  ret = sock->ops->sendmsg(iocb, sock, msg, 16);
#endif
  set_fs(old_fs);

  /* Prepare for sending actual data */
  msg->msg_iov = old_iovec;
  msg->msg_iovlen = n_iovec;

  if (ret == 16)
#if (MAJOR_MINOR_IS(2,4))
    return system_inet_stream_ops->sendmsg(sock, msg, size, scm);
#else /* then it's a 2.6 kernel */
    return system_inet_stream_ops->sendmsg(iocb, sock, msg, size);
#endif
  else
    return ret;
}

int drcom_inet_stream_ops_connect(struct socket *sock, struct sockaddr *uaddr, int addr_len, int flags)
{
  /* Try to connect, before checking for anything else, because:
     - we can be (quite) sure that uaddr contains the destination ip addr, and
     - if connection fails, we just return without checking (but how do we
       check for a connection failure?)
     Maybe we should change this when we can get the dest ip in a sure way */
  int ret = system_inet_stream_ops_connect(sock, uaddr, addr_len, flags);

  if (status_data->status != '1')
    goto normal;

  if (!need_auth(0, ((struct sockaddr_in *) uaddr)->sin_addr.s_addr))
    goto normal;

  /* Why replace sock->ops? Because sock->ops is originally &inet_stream_ops,
     and if we replace just sock->ops->sendmsg, we'll be in trouble if
     we don't recheck the address in sendmsg(). But checking the address for
     every sendmsg() call is very expensive, compared to checking it just in
     connect() */
  sock->ops = &drcom_inet_stream_ops;

  normal:
  return ret;
}

int drcom_udp_get_addrs(u_int32_t *saddr, u_int32_t *daddr, struct sock *sk, struct msghdr *msg)
{
  struct sockaddr_in *usin;
#if (MAJOR_MINOR_IS(2,4))
#else /* then it's a 2.6 kernel */
# if (REVISION < 11) /* 2.6.11 changes a lot of things */
  struct inet_opt *inet = inet_sk(sk);
# else
  struct inet_sock *inet = inet_sk(sk);
# endif
#endif

  /* Get the address */
  /* Stolen from linux-2.4.30/net/ipv4/udp.c */
  /* ... but modified, of course */

  if (msg->msg_name)
  {
    usin = msg->msg_name;
    if (msg->msg_namelen != sizeof(struct sockaddr_in)
        || (usin->sin_family != AF_INET && usin->sin_family != AF_UNSPEC)
        || usin->sin_port == 0)
      return -EINVAL;

    *daddr = usin->sin_addr.s_addr;
  }
  else
  {
#if (MAJOR_MINOR_IS(2,4))
    if (sk->state != TCP_ESTABLISHED)
#else /* then it's a 2.6 kernel */
    if (sk->sk_state != TCP_ESTABLISHED)
#endif
      return -EDESTADDRREQ;

#if (MAJOR_MINOR_IS(2,4))
    *daddr = sk->daddr;
#else /* then it's a 2.6 kernel */
    *daddr = inet->daddr;
#endif
  }

#if (MAJOR_MINOR_IS(2,4))
  *saddr = sk->saddr;
#else /* then it's a 2.6 kernel */
  *saddr = inet->saddr;
#endif

  return 0;
}

#if (MAJOR_MINOR_IS(2,4))
int drcom_inet_dgram_ops_sendmsg(struct socket *sock, struct msghdr *msg, int size, struct scm_cookie *scm)
#else /* then it's a 2.6 kernel */
int drcom_inet_dgram_ops_sendmsg(struct kiocb *iocb, struct socket *sock, struct msghdr *msg, size_t size)
#endif
{
  u_int32_t saddr, daddr;
  struct iovec auth_iovec1[2], *auth_iovec, *old_iovec;
  mm_segment_t old_fs;
  int ret, r;
  unsigned int i;

  if (status_data->status != '1')
    goto normal;

  /* FIXME: Something's wrong here... */
  r = drcom_udp_get_addrs(&saddr, &daddr, sock->sk, msg);
  if (r)
    return r;

  if (!need_auth(saddr, daddr))
    goto normal;

  /* Does any packet have zero iovecs? */
  if (msg->msg_iovlen == 0)
    goto normal;

  /* Save the original iovec pointer */
  old_iovec = msg->msg_iov;

  /* Optimize: usually, msg_iovlen is 1 */
  if (msg->msg_iovlen == 1)
  {
    auth_iovec1[0].iov_base = auth_data->auth;
    auth_iovec1[0].iov_len = 16;
    auth_iovec1[1].iov_base = msg->msg_iov[0].iov_base;
    auth_iovec1[1].iov_len = msg->msg_iov[0].iov_len;
    msg->msg_iov = auth_iovec1;
  }
  else
  {
#ifdef DEBUG
  printk("msg_iovlen: %d\n", msg->msg_iovlen);
#endif
    auth_iovec = kmalloc((msg->msg_iovlen+1) * sizeof(struct iovec), GFP_USER);

    auth_iovec[0].iov_base = auth_data->auth;
    auth_iovec[0].iov_len = 16;

    for (i = 0; i < msg->msg_iovlen; ++i)
    {
      auth_iovec[i+1].iov_base = msg->msg_iov[i].iov_base;
      auth_iovec[i+1].iov_len = msg->msg_iov[i].iov_len;
    }

    msg->msg_iov = auth_iovec;
  }

  /* Now that we have appended the authentication data, increment the
     byte counts so that the packet will be sent correctly */
  ++(msg->msg_iovlen);
  size += 16;

  /* This was the missing piece -- by using set_fs(KERNEL_DS),
     we bypass the check in copy_from_user(), which is called later on
     down the stack */
  old_fs = get_fs();
  set_fs(KERNEL_DS);
#if (MAJOR_MINOR_IS(2,4))
  ret = system_inet_dgram_ops_sendmsg(sock, msg, size, scm);
#else /* then it's a 2.6 kernel */
  ret = system_inet_dgram_ops_sendmsg(iocb, sock, msg, size);
#endif
  set_fs(old_fs);

  /* Restore the original iovec array and the byte counts */
  msg->msg_iov = old_iovec;
  --(msg->msg_iovlen);
  if (ret >= 16)
    ret -= 16;

  return ret;

  normal:
#if (MAJOR_MINOR_IS(2,4))
  return system_inet_dgram_ops_sendmsg(sock, msg, size, scm);
#else /* then it's a 2.6 kernel */
  return system_inet_dgram_ops_sendmsg(iocb, sock, msg, size);
#endif
}

int init_proto(void)
{
#if 0
  get_proc_bufs(&status_data, &auth_data, &iface_data, &except_data);
#endif

  system_inet_stream_ops_connect = inet_stream_ops.connect;
  inet_stream_ops.connect = drcom_inet_stream_ops_connect;

  system_inet_dgram_ops_sendmsg = inet_dgram_ops.sendmsg;
  inet_dgram_ops.sendmsg = drcom_inet_dgram_ops_sendmsg;

  /* drcom_inet_stream_ops is copied to sock->ops after a connection is made */
  memcpy(&drcom_inet_stream_ops, system_inet_stream_ops,
           sizeof(struct proto_ops));
  drcom_inet_stream_ops.connect = drcom_inet_stream_ops_connect;
  drcom_inet_stream_ops.sendmsg = drcom_inet_stream_ops_sendmsg;

  return 0;
}

int cleanup_proto(void)
{
  inet_dgram_ops.sendmsg = system_inet_dgram_ops_sendmsg;
  inet_stream_ops.connect = system_inet_stream_ops_connect;

  return 0;
}

int need_auth(u_int32_t saddr, u_int32_t daddr)
{
  int i;

  /* Check source */
  if (saddr != 0 && saddr != iface_data->hostip)
    goto normal;

  /* Check destination */

  /* DNS servers */
  if ((daddr == iface_data->dnsp) || (daddr == iface_data->dnss))
    goto normal;

  /* Internal network */
  if (
#include "except/private.c"
     ||
#include "except/common.c"
     )
    goto normal;

  /* Exceptions */
  /* Remember to check for both addr and mask! */
  i = 0;
  while ((except_data[i].addr != 0)
         && (except_data[i].mask != 0))
    if ((daddr & except_data[i].mask)
        == (except_data[i].addr & except_data[i].mask))
      goto normal;
    else
      ++i;

  return 1;

  normal:
  return 0;
}

